import typing as t
from numbers import Number
from .base import BaseSchema

__all__ = ('NumberSchema',)

class NumberSchema(BaseSchema):
    def __init__(self, message: t.Optional[str] = None):
        super().__init__((int, float), message=message) #type: ignore

    def float(self):
        self._type = float
        return self

    def int(self):
        self._type = int
        return self

    def lt(self, num: t.Type[Number], message: t.Optional[str] = None):
        def validator(data, schema):
            label = schema._label

            if not label:
                label = 'data'

            nonlocal message
            if not message:
                message = '{label} must be less than {num}'

            if not data < num:
                return message.format(label=label, num=num)
                
        return self.add_validator('is-lt', validator, False)

    def gt(self, num: t.Type[Number], message: t.Optional[str] = None):
        def validator(data, schema):
            label = schema._label

            if not label:
                label = 'data'

            nonlocal message
            if not message:
                message = '{label} must be greater than {num}'

            if not data > num:
                return message.format(label=label, num=num)
                
        return self.add_validator('is-gt', validator, False)

    def lte(self, num: t.Type[Number], message: t.Optional[str] = None):
        def validator(data, schema):
            label = schema._label

            if not label:
                label = 'data'

            nonlocal message
            if not message:
                message = '{label} must be less than or equal to {num}'

            if not data <= num:
                return message.format(label=label, num=num)
                
        return self.add_validator('is-lte', validator, False)

    def gte(self, num: t.Type[Number], message: t.Optional[str] = None):
        def validator(data, schema):
            label = schema._label

            if not label:
                label = 'data'

            nonlocal message
            if not message:
                message = '{label} must be greater than or equal to {num}'

            if not data >= num:
                return message.format(label=label, num=num)
                
        return self.add_validator('is-gte', validator, False)

    def min(self, num: t.Type[Number], message: t.Optional[str] = None):
        return self.gte(num, message)

    def max(self, num: t.Type[Number], message: t.Optional[str] = None):
        return self.lte(num, message)

    def positive(self, message: t.Optional[str] = None):
        return self.gt(0, message) #type: ignore

    def negative(self, message: t.Optional[str] = None):
        return self.lt(0, message) #type: ignore

    def nonpositive(self, message: t.Optional[str] = None):
        return self.lte(0, message) #type: ignore

    def nonnegative(self, message: t.Optional[str] = None):
        return self.gte(0, message) #type: ignore



import copy
import typing as t
from functools import reduce
from dataclasses import dataclass
from .exceptions import ValidationError, ParseError

__all__ = ('Validator', 'BaseSchema')

@dataclass
class Validator:
    name: str
    fn: t.Callable[
        [
            t.Any,
            t.Type['BaseSchema']
        ],
        t.Optional[str]
    ]

    def __call__(self, data: t.Any, schema: t.Type['BaseSchema']):
        return self.fn(data, schema)
        
def def_type_validator(message):
    def wrap(data, schema):
        if not isinstance(data, schema._type):
            types = schema._type
            label = schema._label
            
            if not isinstance(types, tuple):
                types = (types,)

            if not label:
                label = 'data'

            nonlocal message
            if not message:
                message = '{label} type is not of expected type(s): {types}'
            
            t_str = ', '.join(map(lambda t: t.__name__, types))
            return message.format(label=label, types=t_str)
    return wrap

class BaseSchema:
    def __init__(
        self,
        type: t.Union[t.Tuple[t.Callable], t.Callable],
        type_validator: t.Optional[t.Callable[[t.Any, t.Type['BaseSchema']], t.Optional[str]]] = None,
        message: t.Optional[str] = None
    ):
        if not type_validator:
            type_validator = def_type_validator(message)
            
        self._type = type
        self._validators: t.List['Validator'] = [Validator('is-type', type_validator)]
        self._throw: bool = False
        self._errors: t.List[str] = []
        self._label: str = ''
        self._mutations: t.List[t.Callable[[t.Any], t.Any]] = []

        self._is_required = True
        self._required_error = '{label} is a required field'
        self._default_value = None

    def __repr__(self):
        return f'<{self.__class__.__name__} of kohi>'

    __str__ = __repr__

    @property
    def errors(self):
        return self._errors

    def _handle_errors(self):
        if not self._throw or len(self._errors) < 1:
            return
            
        raise ValidationError(self.errors)

    def add_validator(
        self,
        name: str,
        fn: t.Callable[[t.Any, t.Type['BaseSchema']], t.Optional[str]],
        exclusive: bool = True
    ):
        if exclusive and any(map(lambda el: el.name == name, self._validators)):
            return
        self._validators.append(Validator(name=name, fn=fn))
        return self

    def _run_validators(self, data: t.Any):
        for validator in self._validators:
            error = validator(data, self) # type: ignore
            if error:
                self._errors.append(error)

    def _return_or_default(self, data: t.Any):
        if not data is None:
            return data

        if self._is_required and self._default_value is None:
            error = self._required_error.format(label=self._label)
            self._errors.insert(0, error)
            return

        return copy.deepcopy(self._default_value)

    def _validate(self, data: t.Any, checks=True):
        if checks and (data := self._return_or_default(data)) != None:
            self._run_validators(data)
        self._handle_errors()
        return len(self.errors) == 0

    def validate(self, data: t.Any):
        """Validate the given data against the schema"""
        self.reset()
        return self._validate(data)

    def parse(self, data: t.Any):
        """Analyzes the data and returns after passing the validation step"""
        cloned = copy.deepcopy(self._return_or_default(data))

        if (cloned != data and data == None) or (cloned == data == None and not self._is_required):
            return cloned

        try:
            self.throw()._validate(cloned, False)
        except Exception as e:
            raise ParseError(str(e)) from e

        if self._mutations:
            # compose
            mutation = reduce(lambda a, b: lambda c: a(b(c)), reversed(self._mutations), lambda s: s)
            cloned = mutation(cloned)

        return cloned

    def add_mutation(self, mutation: t.Callable[[t.Any], t.Any]):
        self._mutations.append(mutation)
        return self

    def reset(self):
        self._errors = []
    
    def throw(self):
        self._throw = True
        return self

    def label(self, text: str):
        self._label = text
        return self

    def default(self, data: t.Any):
        self._default_value = data
        return self

    def optional(self):
        self._is_required = False    
        return self

    def required(self, message: str='{label} is a required field'):
        self._is_required = True
        self._required_error = message
        return self
        


import re, uuid
import typing as t
from .base import BaseSchema

__all__ = ('StringSchema',)


class StringSchema(BaseSchema):
    def __init__(self, message: t.Optional[str] = None):
        super().__init__((str,), message=message)

    def min(self, min_length: int, message: t.Optional[str] = None):
        def validator(data, schema):
            label = schema._label

            if not label:
                label = 'data'

            nonlocal message
            if not message:
                message = '{label} length must be greater than or equal to {length}'
                
            if len(data) < min_length:
                return message.format(length=min_length, label=label)
                
        return self.add_validator('min-length', validator)
    
    def length(self, length: int, message: t.Optional[str] = None):
        def validator(data, schema):
            label = schema._label

            if not label:
                label = 'data'

            nonlocal message
            if not message:
                message = '{label} length must be equal to {length}'
                
            if len(data) != length:
                return message.format(length=length, label=label)
                
        return self.add_validator('length', validator)

    def max(self, max_length: int, message: t.Optional[str] = None):
        def validator(data, schema):
            label = schema._label

            if not label:
                label = 'data'

            nonlocal message
            if not message:
                message = '{label} length must be less than or equal to {length}'
                
            if len(data) > max_length:
                return message.format(length=max_length, label=label)
                
        return self.add_validator('max-length', validator)
    
    def url(self, message: t.Optional[str] = None):
        def validator(data, schema):
            label = schema._label

            if not label:
                label = 'data'

            nonlocal message
            if not message:
                message = '{label} is not a valid URL'
                
            regex = re.compile(
                r'^(?:http|ftp)s?://'  # http:// or https://
                r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
                r'localhost|'  # localhost...
                r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
                r'(?::\d+)?'  # optional port
                r'(?:/?|[/?]\S+)$', re.IGNORECASE)
            
            if not regex.match(data):
                return message.format(label=label)
                
        return self.add_validator('is-url', validator)

    def uuid(self, message: t.Optional[str] = None):
        def validator(data, schema):
            label = schema._label

            if not label:
                label = 'data'

            nonlocal message
            if not message:
                message = '{label} is not a valid UUID'
                
            try:
                uuid.UUID(data)
            except ValueError:
                return message.format(label=label)
        return self.add_validator('is-uuid', validator)

    def starts_with(self, text: str, message: t.Optional[str] = None):
        def validator(data, schema):
            label = schema._label

            if not label:
                label = 'data'

            nonlocal message
            if not message:
                message = '{label} must starts with {text}'
                
            if not data.startswith(text):
                return message.format(text=text, label=label)
                
        return self.add_validator('starts-with', validator)

    def ends_with(self, text: str, message: t.Optional[str] = None):
        def validator(data, schema):
            label = schema._label

            if not label:
                label = 'data'

            nonlocal message
            if not message:
                message = '{label} must end with {text}'
                
            if not data.endswith(text):
                return message.format(text=text, label=label)
                
        return self.add_validator('ends-with', validator)

    # def email(self):
    #     ...



import typing as t


__all__ = ('ValidationError', 'ParseError')


class ValidationError(Exception):
    def __init__(self, errors: t.List[str]):
        self.errors = errors
        message = errors[0]

        if (len_errors := len(errors)) > 1:
            message = f'{len_errors} errors occurred'

        super().__init__(message)

class ParseError(Exception):
    """error thrown when something goes wrong in the `parse()` method"""



__version__ = '0.3.0'

from .string import StringSchema as StringSchema
from .number import NumberSchema as NumberSchema
from .enum import EnumSchema as EnumSchema
from .dictionary import DictSchema as DictSchema



import typing as t
from .base import BaseSchema

__all__ = ('EnumSchema',)


class EnumSchema(BaseSchema):
    def __init__(self, message: t.Optional[str] = None):
        super().__init__((object,), message=message) #type: ignore

    def one_of(self, opts: t.Iterable, message: t.Optional[str] = None):
        def validator(data, schema):
            label = schema._label

            if not label:
                label = 'data'

            nonlocal message
            if not message:
                message = '{label} is not a valid value'

            if not data in opts:
                return message.format(label=label, value=data, options=opts)
                
        return self.add_validator('one-of', validator)

    def not_one_of(self, opts: t.Iterable, message: t.Optional[str] = None):
        def validator(data, schema):
            label = schema._label

            if not label:
                label = 'data'

            nonlocal message
            if not message:
                message = '{label} is not a valid value'

            if data in opts:
                return message.format(label=label, value=data, options=opts)
                
        return self.add_validator('not_one_of', validator)



import typing as t
from .base import BaseSchema


__all__ = ('DictSchema',)


class DictSchema(BaseSchema):
    def __init__(self, message: t.Optional[str] = None):
        super().__init__((dict,), message=message)
        self._props: dict = {}

    def props(self, **props):
        self._props = props
        return self

    def _prepare_new_label(self, key: str):
        if self._label:
            return f'{self._label}.{key}'
        return key

    def _validate_props(self):
        for key, schema in self._props.items():
            if not isinstance(schema, BaseSchema):
                raise TypeError(f'The validator at {key} is not a valid schema validator')

            def validator(data, parent):
                nonlocal schema, key
                validate = schema.label(parent._prepare_new_label(key))

                _data = None
                if key in data:
                    _data = data[key]

                if not validate.validate(_data):
                    new_errors = [*parent.errors, *validate.errors]
                    parent._errors = new_errors

            self.add_validator(f'valide-{key}', validator)

    def _validate(self, data: t.Any, checks=True):
        self._validate_props()
        return super()._validate(data, checks)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kohi']

package_data = \
{'': ['*']}

install_requires = \
['pytest>=7.2.2,<8.0.0']

setup_kwargs = {
    'name': 'kohi',
    'version': '0.4.0',
    'description': 'A powerfull schema validator',
    'long_description': '# kohi\n\n<p align="center">A powerfull schema validator</p>\n\n![GitHub Repo stars](https://img.shields.io/github/stars/natanfeitosa/kohi)\n![GitHub Workflow Status](https://img.shields.io/github/actions/workflow/status/natanfeitosa/kohi/pytest.yml?label=Pytest&logo=github)\n![GitHub](https://img.shields.io/github/license/natanfeitosa/kohi)\n![PyPI - Format](https://img.shields.io/pypi/format/kohi)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/kohi)\n![PyPI - Package Version](https://img.shields.io/pypi/v/kohi)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/kohi)\n[![Open Source Helpers](https://www.codetriage.com/natanfeitosa/kohi/badges/users.svg)](https://www.codetriage.com/natanfeitosa/kohi)\n\n## Instalation\n\nVia Poetry:\n```sh\npoetry add kohi\n```\n\nVia PIP:\n```sh\npip install kohi\n```\n\nVia GitHub (recommended only in dev env):\n```sh\ngit clone https://github.com/natanfeitosa/kohi.git && cd kohi && pip install .\n```\n\n## Quickstart\n\nTo validate a type you can import your schema validator from `kohi` or `from kohi.<type> import <type>Schema`\n\ne.g.\n\nLet\'s check if a person\'s date of birth is a positive integer less than the current date — 2023 — and greater than or equal to 2005\n\n```python\nfrom kohi import NumberSchema\n# from kohi.number import NumberSchema\n\nn = NumberSchema().int().positive().lt(2023).gte(2005)\n\nprint(n.validate(2005)) # True\nprint(n.validate(2022)) # True\nprint(n.validate(2004)) # False\nprint(n.validate(2023)) # False\n\n# You can see the errors generated in the last `validate` call just by accessing the `errors` property\n# print(n.errors) # [\'number must be less than 2022\']\n```\n\n## Validators\n\n* [`kohi.base.BaseSchema`](#baseschema)\n> Only one base class for all schema validators\n* [`kohi.number.NumberSchema`](#numberschema)\n> or `kohi.NumberSchema`\n* [`kohi.string.StringSchema`](#stringschema)\n> or `kohi.StringSchema`\n* [`kohi.enum.EnumSchema`](#enumschema)\n> or `kohi.EnumSchema`\n* [`kohi.dictionary.DictSchema`](#dictschema)\n> or `kohi.DictSchema`\n\n## Methods\n\n### `BaseSchema`\n* `add_validator(name, func): Self`\n  > Add a custom data validator\n* `validate(data): bool`\n  > The method to be called when we validate the schema\n* `reset(): None`\n  > Reset error list\n* `throw(): Self`\n  > By default no errors are thrown, but when this method is chained a `ValidationError` will be thrown\n* `add_mutation(): Self`\n  > Add a mutation function than will run after the `validate` method. P.S. Will only be executed in the `parse` method\n* `parse(data): typeof data`\n  > Run the `validate` method, the mutations and return a deep clone of data\n* `default(data): Self`\n  > Set a default value for when the validator receives None and you don\'t want to generate an error\n* `optional(): Self`\n  > Allow values None\n* `required(error_message=None): Self`\n  > Mark the schema as required. Does not allow values None\n\n### `NumberSchema`\ninherits from [`BaseSchema`](#baseschema)\n> By default validates int and float \n\n* `float(): Self`\n  > Validate only `float`\n* `int(): Self`\n  > Validate only `int`\n* `lt(num): Self`\n  > Validates if the data is less than `num`\n* `gt(num): Self`\n  > Validates if the data is greater than `num`\n* `lte(num): Self`\n  > Validates if the data is less than or equal to `num`\n* `gte(num): Self`\n  > Validates if the data is greater than or equal to `num`\n* `min(num): Self`\n  > Just an alias for `gte(num)`\n* `max(num): Self`\n  > Just an alias for `lte(nun)`\n* `positive(): Self`\n  > Just an alias for `gt(0)`\n* `negative(): Self`\n  > Just an alias for `lt(0)`\n* `nonpositive(): Self`\n  > Just an alias for `lte(0)`\n* `nonnegative(): Self`\n  > Just an alias for `gte(0)`\n\n### StringSchema\ninherits from [`BaseSchema`](#baseschema)\n\n* `min(min_length): Self`\n  > Validate if the data len is greater than or equal to min_length\n* `length(length): Self`\n  > Validate if the data len equal to length\n* `max(max_length): Self`\n  > Validate if the data len is less than or equal to max_length\n* `url(): Self`\n  > Validate if the data is an url\n* `uuid(): Self`\n  > Validate if the data is a valid uuid\n* `starts_with(text): Self`\n  > Validate if the data starts with text\n* `ends_with(text): Self`\n  > Validate if the data ends with text\n\n### EnumSchema\ninherits from [`BaseSchema`](#baseschema)\n\n* `one_of(opts): Self`\n  > Validate if the data is in opts\n* `not_one_of(opts): Self`\n  > Validate that data is different from the values in opts\n\n### DictSchema\ninherits from [`BaseSchema`](#baseschema)\n\n* `props(**props): Self`\n  > Defines the structure of the dictionary in the format `[key]: ClassValidator`\n\n## Dev env\n\n* install development dependencies\n* check types using `mypy`\n* run all tests using `pytest`\n',
    'author': 'Natan Santos',
    'author_email': 'natansantosapps@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/natanfeitosa/kohi#readme',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
